from typing import Tuple, Union

import numpy as np
from numpy.typing import ArrayLike


def unique_rows(
    a: ArrayLike, return_inverse: bool = False, return_counts: bool = False
) -> Union[np.ndarray, Tuple]:
    # The numpy alternative `np.unique(a, axis=0)` is slow; cf.
    # <https://github.com/numpy/numpy/issues/11136>.
    a = np.asarray(a)

    a_shape = a.shape
    a = a.reshape(a.shape[0], np.prod(a.shape[1:], dtype=int))

    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    out = np.unique(b, return_inverse=return_inverse, return_counts=return_counts)
    # out[0] are the sorted, unique rows
    if isinstance(out, tuple):
        out = (out[0].view(a.dtype).reshape(out[0].shape[0], *a_shape[1:]), *out[1:])
    else:
        out = out.view(a.dtype).reshape(out.shape[0], *a_shape[1:])

    return out



import numpy as np
import numpy.typing as npt


# There also is
# <https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.logsumexp.html>,
# but implementation is easy enough
def _logsumexp(x: npt.ArrayLike):
    c = np.max(x)
    return c + np.log(np.sum(np.exp(x - c)))


def mean(x: npt.ArrayLike, p: float = 1):
    x = np.asarray(x)

    n = len(x)
    if p == 1:
        return np.mean(x)
    if p == -np.inf:
        return np.min(np.abs(x))
    elif p == 0:
        # first compute the root, then the product, to avoid numerical difficulties with
        # too small values of prod(x)
        if np.any(x < 0.0):
            raise ValueError("p=0 only works with nonnegative x.")
        return np.prod(np.power(x, 1 / n))
    elif p == np.inf:
        return np.max(np.abs(x))

    if not isinstance(p, int) and np.any(x < 0.0):
        raise ValueError("Non-integer p only work with nonnegative x.")

    if np.all(x > 0.0):
        # logsumexp trick to avoid overflow for large p
        # only works for positive x though
        return np.exp((_logsumexp(p * np.log(x)) - np.log(n)) / p)
    else:
        return (np.sum(x ** p) / n) ** (1.0 / p)
